/*
    ettercap -- data decoding module

    Copyright (C) 2001  ALoR <alor@users.sourceforge.net>, NaGA <crwm@freemail.it>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "include/ec_main.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#ifdef HAVE_CTYPE_H
   #include <ctype.h>
#endif

#include "include/ec_inet_structures.h"
#include "include/ec_error.h"

#ifdef DEBUG
   #include "include/ec_debug.h"
#endif

typedef struct
{
   char ip_source[16];
   char ip_dest[16];
   int source_port;
   int dest_port;
   u_long source_seq;
   u_long dest_seq;
   char flags;
   char status[8];
   struct conn_list *next;
} conn_list;

typedef struct
{
   int port;
   char description[18];
   struct database *next;
} database;

database *d_list = NULL;

conn_list *c_list = NULL;

#ifndef HAVE_CTYPE_H
   int isprint(int c);
#endif
int Decodedata_MakeConnectionList(CONNECTION data);
int Decodedata_RefreshConnectionList(void);
void Decodedata_Free_list(conn_list *head);
void Decodedata_Make_list(CONNECTION data);
char * Decodedata_GetType(int port1, int port2);

char * Decodedata_GetAsciiData(char *buffer, int buff_len);
char * Decodedata_GetHexData(char *buffer, int buff_len, short dimX);
char * Decodedata_GetEnhanchedHexData(char *buffer, int buff_len, short cr);
char * Decodedata_TCPFlags(char flags);

//--------------------------

#ifndef HAVE_CTYPE_H

   int isprint(int c)
   {
      return ( (c>31 && c<127) ? 1 : 0 );
   }

#endif


int Decodedata_MakeConnectionList(CONNECTION data)
{
   int num_conn = 0, i = 0;
   conn_list *c_index;

   Decodedata_Make_list(data);

   c_index = c_list;

   for( ; c_index; num_conn++)
      c_index = (conn_list *)c_index->next;

   if (Conn_Between_Hosts) free(Conn_Between_Hosts);
   if ( (Conn_Between_Hosts = (CONNECTION *)calloc(num_conn, sizeof(CONNECTION))) == NULL)
      Error_msg("ec_decodedata:%d calloc() | ERRNO : %d | %s", __LINE__, errno, sys_errlist[errno]);

   c_index = c_list;

   for( ; c_index; c_index = (conn_list *)c_index->next)
   {
      memcpy(&Conn_Between_Hosts[i].ip_source, &c_index->ip_source, sizeof(c_index->ip_source));
      memcpy(&Conn_Between_Hosts[i].ip_dest, &c_index->ip_dest, sizeof(c_index->ip_dest));
      memcpy(&Conn_Between_Hosts[i].source_port, &c_index->source_port, sizeof(int));
      memcpy(&Conn_Between_Hosts[i].dest_port, &c_index->dest_port, sizeof(int));
      memcpy(&Conn_Between_Hosts[i].source_seq, &c_index->source_seq, sizeof(u_long));
      memcpy(&Conn_Between_Hosts[i].dest_seq, &c_index->dest_seq, sizeof(u_long));

      strcpy(Conn_Between_Hosts[i].status, c_index->status);
      strcpy(Conn_Between_Hosts[i].type, Decodedata_GetType(Conn_Between_Hosts[i].source_port,
                                                            Conn_Between_Hosts[i].dest_port));

      i++;
   }

   return num_conn;

}


int Decodedata_RefreshConnectionList(void)
{

#ifdef DEBUG
   Debug_msg("Decodedata_RefreshConnectionList");
#endif

   if (Conn_Between_Hosts) free(Conn_Between_Hosts);
   Conn_Between_Hosts = NULL;
   Decodedata_Free_list(c_list);
   c_list = NULL;

   return 0;
}



void Decodedata_Free_list(conn_list *head)
{
   if (!head)  return;

   Decodedata_Free_list((conn_list *) head->next);
   free(head);
}



void Decodedata_Make_list(CONNECTION data)
{
   conn_list *node;
   conn_list *c_index = NULL;
   conn_list *temp;
   char filter[45];
   char found = 0;

   sprintf(filter, "%s%d%s%d", data.ip_source, data.source_port, data.ip_dest, data.dest_port);

   temp = c_list;

   while (temp != NULL)
   {
      char cur_conn[45];
      char rev_cur_conn[45];
      sprintf(cur_conn, "%s%d%s%d", temp->ip_source, temp->source_port, temp->ip_dest, temp->dest_port);
      sprintf(rev_cur_conn, "%s%d%s%d", temp->ip_dest, temp->dest_port, temp->ip_source, temp->source_port);

      if (!strcmp(filter, cur_conn))
      {
         temp->source_seq = data.source_seq;
         temp->flags = data.flags;
         found = 1;
      }

      if (!strcmp(filter, rev_cur_conn))
      {
         temp->dest_seq = data.dest_seq;
         temp->flags = data.flags;
         found = 1;
      }

      if (found)
      {
         if (temp->flags & TH_RST)
            strcpy(temp->status,  "KILLED");
         else if (temp->flags & TH_SYN)
            strcpy(temp->status,  "OPENING");
         else if (temp->flags & TH_FIN)
            strcpy(temp->status,  "CLOSING");
         else if (temp->flags & TH_PSH)
            strcpy(temp->status,  "ACTIVE");
         else
            if (!strcmp(temp->status, "CLOSING"))     // FIN ACK
               strcpy(temp->status,  "CLOSED");       // ACK
            else if (strcmp(temp->status, "CLOSED"))
               strcpy(temp->status,  "silent");
      }

      if (temp != NULL) c_index = temp;

      temp = (conn_list *)temp->next;
   }

   if (!found)
   {
      #ifdef DEBUG
         Debug_msg("Decodedata_Make_list - new node ! %s:%d - %s:%d", data.ip_source, data.source_port, data.ip_dest, data.dest_port);
      #endif

      if ( (node = (conn_list *)calloc(1,sizeof(conn_list))) == NULL)
         Error_msg("ec_decodedata:%d calloc() | ERRNO : %d | %s", __LINE__, errno, sys_errlist[errno]);

      node->next=NULL;

      memcpy(&node->ip_source, &data.ip_source, sizeof(data.ip_source));
      memcpy(&node->ip_dest, &data.ip_dest, sizeof(data.ip_dest));
      memcpy(&node->source_port, &data.source_port, sizeof(int));
      memcpy(&node->dest_port, &data.dest_port, sizeof(int));
      memcpy(&node->source_seq, &data.source_seq, sizeof(u_long));
      memcpy(&node->dest_seq, &data.dest_seq, sizeof(u_long));
      memcpy(&node->flags, &data.flags, sizeof(char));

         if (node->flags & TH_RST)
            strcpy(node->status,  "KILLED");
         else if (node->flags & TH_FIN)
            strcpy(node->status,  "CLOSING");
         else if (node->flags & TH_SYN)
            strcpy(node->status,  "OPENING");
         else if (node->flags & TH_PSH)
            strcpy(node->status,  "ACTIVE");
         else
            strcpy(node->status,  "silent");

      if (c_index != NULL)
         (conn_list *)c_index->next = node;

      if (c_list == NULL) c_list = node;
   }

}



char * Decodedata_GetHexData(char *buffer, int buff_len, short dimX)
{
   short octets;

   for(octets = 0; octets < dimX; octets++)
   {
      if ( (octets*3.5 + 12) >= dimX ) break;
   }

   if (octets > 16) octets = 16;
   if (octets % 2 == 1) octets--;

   return Decodedata_GetEnhanchedHexData(buffer, buff_len, octets);

}


char * Decodedata_GetEnhanchedHexData(char *buffer, int buff_len, short cr)
{
   static char *hexdata;
   int i, j, jm;
   int c, dim = 0;

   if (buff_len == 0) return "";

   c = cr*3.5 + 11;
   dim = c;

   for (i = 0; i < buff_len; i++)   // approximately
      if ( i % cr == 0)             // approximately
         dim += c;                  // approximately


   if (hexdata) free(hexdata);
   if ( (hexdata = (char *)calloc(dim, sizeof(char))) == NULL)
      Error_msg("ec_decodedata:%d calloc() | ERRNO : %d | %s", __LINE__, errno, sys_errlist[errno]);

   // adapted from dsniff by Dug Song <dugsong@monkey.org>
   sprintf(hexdata,"\n");
   for (i = 0; i < buff_len; i += cr) {
           sprintf(hexdata, "%s %04x: ", hexdata, i );
           jm = buff_len - i;
           jm = jm > cr ? cr : jm;

           for (j = 0; j < jm; j++) {
                   if ((j % 2) == 1) sprintf(hexdata,"%s%02x ", hexdata, (unsigned char) buffer[i+j]);
                   else sprintf(hexdata,"%s%02x", hexdata, (unsigned char) buffer[i+j]);
           }
           for (; j < cr; j++) {
                   if ((j % 2) == 1) strcat(hexdata,"   ");
                   else strcat(hexdata,"  ");
           }
           strcat(hexdata," ");

           for (j = 0; j < jm; j++) {
                   c = buffer[i+j];
                   c = isprint(c) ? c : '.';
                   sprintf(hexdata,"%s%c", hexdata, c);
           }
           strcat(hexdata,"\n");
   }

   return hexdata;
}



char * Decodedata_GetAsciiData(char *buffer, int buff_len)
{

   int i = 0;

   if (buff_len == 0) return "";

   for(i = 0; i < buff_len; i++)
   {
      if ( !( isprint((int)buffer[i]) || buffer[i] == '\n' || buffer[i] == '\t') )
         buffer[i] = '.';
   }

   return buffer;
}


char * Decodedata_TCPFlags(char flags)
{
   static char string[8];
   char *p;

   memset(string, 0, 8);
   p = string;

   if (flags & TH_ACK) *p++ = 'A';
   if (flags & TH_PSH) *p++ = 'P';
   if (flags & TH_RST) *p++ = 'R';
   if (flags & TH_SYN) *p++ = 'S';
   if (flags & TH_FIN) *p++ = 'F';

   return string;

}


char * Decodedata_GetType(int port1, int port2)
{

   static char type[18];
   database *d_index;

   if (d_list == NULL)  // only the first time
   {
      FILE *f_ser;
      char line[1024], desc[18], stype[4];
      int port;

#ifdef DEBUG
   Debug_msg("Decodedata_GetType - loading from /etc/services");
#endif

      if ( (d_index = (database *)calloc(1,sizeof(database))) == NULL)
         Error_msg("ec_decodedata:%d calloc() | ERRNO : %d | %s", __LINE__, errno, sys_errlist[errno]);

      d_list = d_index;

      if (!(f_ser = fopen ("/etc/services", "r")))
         Error_msg("ec_decodedata:%d fopen(\"/etc/services\") | ERRNO : %d | %s", __LINE__, errno, sys_errlist[errno]);

      while (fgets (line, 1024, f_ser))
      {
         if ((sscanf (line, "%16s%u/%s", desc, &port, stype) == 3) && (!strstr (desc, "#")) && (strstr (stype, "tcp")))
         {
            if ( (d_index->next = ( struct database *) calloc (1, sizeof(database))) == NULL)
               Error_msg("ec_decodedata:%d calloc() | ERRNO : %d | %s", __LINE__, errno, sys_errlist[errno]);
            d_index->port = port;
            strcpy (d_index->description, desc);
            d_index = (database *) d_index->next;
         }
      }

      fclose (f_ser);
      d_index->next = NULL;
   }

   d_index = d_list;
   for( ; d_index; d_index = (database *)d_index->next)
   {
      if ( port1 == d_index->port || port2 == d_index->port )
      {
         strcpy(type, d_index->description);
         return type;
      }
   }

   return "";
}

/* EOF */
